# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Define few unrelated algorithms that work on declarations.

"""

from . import declaration_utils
from . import runtime_errors


class match_declaration_t(object):
    """
    Helper class for different search algorithms.

    This class will help developer to match declaration by:
        - declaration type, for example :class:`class_t` or
            :class:`operator_t`.
        - declaration name
        - declaration full name
        - reference to parent declaration

    """

    def __init__(
            self, decl_type=None,
            name=None, fullname=None, parent=None):

        self._decl_type = decl_type
        self.name = name
        self.fullname = fullname
        self.parent = parent

    def does_match_exist(self, inst):
        """
        Returns True if inst does match one of specified criteria.

        :param inst: declaration instance
        :type inst: :class:`declaration_t`

        :rtype: bool

        """

        answer = True
        if self._decl_type is not None:
            answer &= isinstance(inst, self._decl_type)
        if self.name is not None:
            answer &= inst.name == self.name
        if self.parent is not None:
            answer &= self.parent is inst.parent
        if self.fullname is not None:
            if inst.name:
                answer &= self.fullname == declaration_utils.full_name(inst)
            else:
                answer = False
        return answer

    def __call__(self, inst):
        """
        .. code-block:: python

           return self.does_match_exist(inst)

        """

        return self.does_match_exist(inst)


def apply_visitor(visitor, decl_inst):
    """
    Applies a visitor on declaration instance.

    :param visitor: instance
    :type visitor: :class:`type_visitor_t` or :class:`decl_visitor_t`

    """

    fname = 'visit_' + \
        decl_inst.__class__.__name__[:-2]  # removing '_t' from class name
    if not hasattr(visitor, fname):
        raise runtime_errors.visit_function_has_not_been_found_t(
            visitor, decl_inst)
    return getattr(visitor, fname)()



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name,
    castxml_epic_version=1)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)

a1 = global_namespace.variable("a1")
print(str(a1.decl_type), type(a1.decl_type))
# > 'A', <class 'pygccxml.declarations.cpptypes.declarated_t'>

print(declarations.is_elaborated(a1.decl_type))
# > False

a2 = global_namespace.variable("a2")
print(str(a2.decl_type), type(a2.decl_type))
# > 'class ::A', <class 'pygccxml.declarations.cpptypes.elaborated_t'>

print(declarations.is_elaborated(a2.decl_type))
# > True

base = declarations.remove_elaborated(a2.decl_type)
print(str(base), type(base))
# > 'A', <class 'pygccxml.declarations.cpptypes.declarated_t'>

# The same can be done with function arguments:
fun = global_namespace.free_function("function")
print(type(fun.arguments[0].decl_type), type(fun.arguments[1].decl_type))
# > <class 'pygccxml.declarations.cpptypes.declarated_t'>,
#   <class 'pygccxml.declarations.cpptypes.elaborated_t'>



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines classes, that describes C++ classes

This modules contains definition for next C++ declarations:
    - class definition
    - class declaration
    - small helper class for describing C++ class hierarchy
"""

from . import scopedef
from . import declaration_utils
from . import declaration
from . import templates
from . import byte_info
from . import elaborated_info
from . import type_traits


class ACCESS_TYPES(object):

    """class that defines "access" constants"""
    PUBLIC = "public"
    PRIVATE = "private"
    PROTECTED = "protected"
    ALL = [PUBLIC, PRIVATE, PROTECTED]


class CLASS_TYPES(object):

    """class that defines "class" type constants"""
    CLASS = "class"
    STRUCT = "struct"
    UNION = "union"
    ALL = [CLASS, STRUCT, UNION]


def get_partial_name(name):
    from . import container_traits  # prevent cyclic dependencies
    ct = container_traits.find_container_traits(name)

    if ct:
        return ct.remove_defaults(name)

    if templates.is_instantiation(name):
        tmpl_name, args = templates.split(name)
        for i, arg_name in enumerate(args):
            args[i] = get_partial_name(arg_name.strip())
        return templates.join(tmpl_name, args)

    return name


class hierarchy_info_t(object):

    """describes class relationship"""

    def __init__(self, related_class=None, access=None, is_virtual=False):
        """creates class that contains partial information about class
        relationship"""
        if related_class:
            assert isinstance(related_class, class_t)
        self._related_class = related_class
        if access:
            assert access in ACCESS_TYPES.ALL
        self._access = access
        self._is_virtual = is_virtual
        self._declaration_path = None
        self._declaration_path_hash = None

    def __eq__(self, other):
        if not isinstance(other, hierarchy_info_t):
            return False
        return (self.declaration_path_hash ==
                other.declaration_path_hash) \
            and self._declaration_path == other._declaration_path \
            and self._access == other._access \
            and self._is_virtual == other._is_virtual

    def __hash__(self):
        return self.declaration_path_hash

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return self.__class__.__name__ < other.__class__.__name__
        return (self.declaration_path, self.access, self.is_virtual) < \
            (other.declaration_path, other.access, other.is_virtual)

    @property
    def related_class(self):
        """reference to base or derived :class:`class <class_t>`"""
        return self._related_class

    @related_class.setter
    def related_class(self, new_related_class):
        if new_related_class:
            assert isinstance(new_related_class, class_t)
        self._related_class = new_related_class
        self._declaration_path = None
        self._declaration_path_hash = None

    @property
    def access(self):
        return self._access

    @access.setter
    def access(self, new_access):
        assert new_access in ACCESS_TYPES.ALL
        self._access = new_access

    # TODO: Why is there an access_type / access which are the same ?
    @property
    def access_type(self):
        """describes :class:`hierarchy type <ACCESS_TYPES>`"""
        return self.access

    @access_type.setter
    def access_type(self, new_access_type):
        self.access = new_access_type

    # TODO: check whether GCC XML support this and if so parser this
    # information
    @property
    def is_virtual(self):
        """indicates whether the inheritance is virtual or not"""
        return self._is_virtual

    @is_virtual.setter
    def is_virtual(self, new_is_virtual):
        self._is_virtual = new_is_virtual

    @property
    def declaration_path(self):
        if self._declaration_path is None:
            self._declaration_path = declaration_utils.declaration_path(
                self.related_class)
        return self._declaration_path

    @property
    def declaration_path_hash(self):
        if self._declaration_path_hash is None:
            self._declaration_path_hash = hash(tuple(self.declaration_path))
        return self._declaration_path_hash


class class_declaration_t(declaration.declaration_t):

    """describes class declaration"""

    def __init__(self, name=''):
        """creates class that describes C++ class declaration
        ( and not definition )"""
        declaration.declaration_t.__init__(self, name)
        self._aliases = []

    def _get__cmp__items(self):
        """implementation details"""
        return []

    def i_depend_on_them(self, recursive=True):
        self._warn_deprecated()
        return []

    @property
    def aliases(self):
        """List of :class:`aliases <typedef_t>` to this instance"""
        return self._aliases

    @aliases.setter
    def aliases(self, new_aliases):
        self._aliases = new_aliases

    def _get_partial_name_impl(self):
        return get_partial_name(self.name)


class class_t(
        scopedef.scopedef_t,
        byte_info.byte_info,
        elaborated_info.elaborated_info):

    """describes class definition"""

    def __init__(
            self,
            name='',
            class_type=CLASS_TYPES.CLASS,
            is_abstract=False):
        """creates class that describes C++ class definition"""
        scopedef.scopedef_t.__init__(self, name)
        byte_info.byte_info.__init__(self)
        elaborated_info.elaborated_info.__init__(self, class_type)
        if class_type:
            assert class_type in CLASS_TYPES.ALL
        self._class_type = class_type
        self._bases = []
        self._derived = []
        self._is_abstract = is_abstract
        self._public_members = []
        self._private_members = []
        self._protected_members = []
        self._aliases = []
        self._recursive_bases = None
        self._recursive_derived = None

    def _get_name_impl(self):
        return self._name

    def __str__(self):
        name = declaration_utils.full_name(self)
        if name[:2] == "::":
            name = name[2:]
        return "%s [%s]" % (name, self.class_type)

    def _get__cmp__scope_items(self):
        """implementation details"""
        return [
            self.class_type,
            [declaration_utils.declaration_path(base.related_class) for
             base in self.bases].sort(),
            [declaration_utils.declaration_path(derive.related_class) for
             derive in self.derived].sort(),
            self.is_abstract,
            self.public_members.sort(),
            self.private_members.sort(),
            self.protected_members.sort()]

    def __eq__(self, other):
        if not scopedef.scopedef_t.__eq__(self, other):
            return False
        return self.class_type == other.class_type \
            and [declaration_utils.declaration_path(base.related_class) for
                 base in self.bases].sort() \
            == [declaration_utils.declaration_path(base.related_class) for
                base in other.bases].sort() \
            and [declaration_utils.declaration_path(derive.related_class) for
                 derive in self.derived].sort() \
            == [declaration_utils.declaration_path(derive.related_class) for
                derive in other.derived].sort() \
            and self.is_abstract == other.is_abstract \
            and self.public_members.sort() \
            == other.public_members.sort() \
            and self.private_members.sort() \
            == other.private_members.sort() \
            and self.protected_members.sort() \
            == other.protected_members.sort()

    def __hash__(self):
        return hash(self.class_type)

    @property
    def class_type(self):
        """describes class :class:`type <CLASS_TYPES>`"""
        return self._class_type

    @class_type.setter
    def class_type(self, new_class_type):
        if new_class_type:
            assert new_class_type in CLASS_TYPES.ALL
        self._class_type = new_class_type

    @property
    def bases(self):
        """list of :class:`base classes <hierarchy_info_t>`"""
        return self._bases

    @bases.setter
    def bases(self, new_bases):
        self._bases = new_bases

    @property
    def recursive_bases(self):
        """list of all :class:`base classes <hierarchy_info_t>`"""
        if self._recursive_bases is None:
            to_go = self.bases[:]
            all_bases = []
            while to_go:
                base = to_go.pop()
                if base not in all_bases:
                    all_bases.append(base)
                    to_go.extend(base.related_class.bases)
            self._recursive_bases = all_bases
        return self._recursive_bases

    @property
    def derived(self):
        """list of :class:`derived classes <hierarchy_info_t>`"""
        return self._derived

    @derived.setter
    def derived(self, new_derived):
        self._derived = new_derived

    @property
    def recursive_derived(self):
        """list of all :class:`derive classes <hierarchy_info_t>`"""
        if self._recursive_derived is None:
            to_go = self.derived[:]
            all_derived = []
            while to_go:
                derive = to_go.pop()
                if derive not in all_derived:
                    all_derived.append(derive)
                    to_go.extend(derive.related_class.derived)
            self._recursive_derived = all_derived
        return self._recursive_derived

    @property
    def is_abstract(self):
        """describes whether class abstract or not"""
        return self._is_abstract

    @is_abstract.setter
    def is_abstract(self, is_abstract):
        self._is_abstract = is_abstract

    @property
    def public_members(self):
        """list of all public :class:`members <declarationt_>`"""
        return self._public_members

    @public_members.setter
    def public_members(self, new_public_members):
        self._public_members = new_public_members

    @property
    def private_members(self):
        """list of all private :class:`members <declarationt_>`"""
        return self._private_members

    @private_members.setter
    def private_members(self, new_private_members):
        self._private_members = new_private_members

    @property
    def protected_members(self):
        """list of all protected :class:`members <declarationt_>`"""
        return self._protected_members

    @protected_members.setter
    def protected_members(self, new_protected_members):
        self._protected_members = new_protected_members

    @property
    def aliases(self):
        """List of :class:`aliases <typedef_t>` to this instance"""
        return self._aliases

    @aliases.setter
    def aliases(self, new_aliases):
        self._aliases = new_aliases

    def _get_declarations_impl(self):
        return self.get_members()

    def get_members(self, access=None):
        """
        returns list of members according to access type

        If access equals to None, then returned list will contain all members.
        You should not modify the list content, otherwise different
        optimization data will stop work and may to give you wrong results.

        :param access: describes desired members
        :type access: :class:ACCESS_TYPES

        :rtype: [ members ]
        """
        if access == ACCESS_TYPES.PUBLIC:
            return self.public_members
        elif access == ACCESS_TYPES.PROTECTED:
            return self.protected_members
        elif access == ACCESS_TYPES.PRIVATE:
            return self.private_members

        all_members = []
        all_members.extend(self.public_members)
        all_members.extend(self.protected_members)
        all_members.extend(self.private_members)
        return all_members

    def adopt_declaration(self, decl, access):
        """adds new declaration to the class

        :param decl: reference to a :class:`declaration_t`

        :param access: member access type
        :type access: :class:ACCESS_TYPES
        """
        if access == ACCESS_TYPES.PUBLIC:
            self.public_members.append(decl)
        elif access == ACCESS_TYPES.PROTECTED:
            self.protected_members.append(decl)
        elif access == ACCESS_TYPES.PRIVATE:
            self.private_members.append(decl)
        else:
            raise RuntimeError("Invalid access type: %s." % access)
        decl.parent = self
        decl.cache.reset()
        decl.cache.access_type = access

    def remove_declaration(self, decl):
        """
        removes decl from  members list

        :param decl: declaration to be removed
        :type decl: :class:`declaration_t`
        """

        access_type = self.find_out_member_access_type(decl)
        if access_type == ACCESS_TYPES.PUBLIC:
            container = self.public_members
        elif access_type == ACCESS_TYPES.PROTECTED:
            container = self.protected_members
        else:  # decl.cache.access_type == ACCESS_TYPES.PRVATE
            container = self.private_members
        del container[container.index(decl)]
        decl.cache.reset()

    def find_out_member_access_type(self, member):
        """
        returns member access type

        :param member: member of the class
        :type member: :class:`declaration_t`

        :rtype: :class:ACCESS_TYPES
        """
        assert member.parent is self
        if not member.cache.access_type:
            if member in self.public_members:
                access_type = ACCESS_TYPES.PUBLIC
            elif member in self.protected_members:
                access_type = ACCESS_TYPES.PROTECTED
            elif member in self.private_members:
                access_type = ACCESS_TYPES.PRIVATE
            else:
                raise RuntimeError(
                    "Unable to find member within internal members list.")
            member.cache.access_type = access_type
            return access_type
        else:
            return member.cache.access_type

    def __find_out_member_dependencies(self, access_type):
        members = self.get_members(access_type)
        answer = []
        for mem in members:
            answer.extend(mem.i_depend_on_them(recursive=True))
        member_ids = set([id(m) for m in members])
        for dependency in answer:
            if id(dependency.declaration) in member_ids:
                dependency.access_type = access_type
        return answer

    def i_depend_on_them(self, recursive=True):
        self._warn_deprecated()
        # Deprecated method. The cyclic import will be removed with the method
        # in the next release, so we can disable the cyclic import check here.
        from . import dependencies  # pylint: disable=R0401

        answer = []

        for base in self.bases:
            answer.append(
                dependencies.dependency_info_t(
                    self,
                    base.related_class,
                    base.access_type,
                    "base class"))

        if recursive:
            for access_type in ACCESS_TYPES.ALL:
                answer.extend(self.__find_out_member_dependencies(access_type))

        return answer

    def _get_partial_name_impl(self):
        if type_traits.is_std_string(self):
            return 'string'
        if type_traits.is_std_wstring(self):
            return 'wstring'

        return get_partial_name(self.name)

    @property
    def top_class(self):
        """reference to a parent class, which contains this class and defined
        within a namespace

        if this class is defined under a namespace, self will be returned"""
        curr = self
        parent = self.parent
        while isinstance(parent, class_t):
            curr = parent
            parent = parent.parent
        return curr


class_types = (class_t, class_declaration_t)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = ['free_operators.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_free_operators(global_ns):
    fo = global_ns.namespace('free_operators')
    number = fo.class_('number')
    rational = fo.class_('rational')
    for oper in fo.free_operators():
        if number.name in str(oper):
            assert number in oper.class_types
        if rational.name in str(oper):
            assert rational in oper.class_types



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "classes.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_global(global_ns):
    global_ns.class_('cls')
    global_ns.class_('::cls')


def test_typedefs(global_ns):
    global_ns.class_('cls2')
    global_ns.typedef('cls2')
    global_ns.class_('::cls2')

    global_ns.class_('cls3')
    global_ns.typedef('cls3')
    cls3 = global_ns.class_('::cls3')
    cls3.variable('i')


def test_ns1(global_ns):
    ns1 = global_ns.namespace('ns')

    global_ns.class_('nested_cls')
    with pytest.raises(Exception):
        global_ns.class_('ns::nested_cls')
    global_ns.class_('::ns::nested_cls')

    with pytest.raises(Exception):
        ns1.class_('::nested_cls')
    ns1.class_('nested_cls')
    ns1.class_('::ns::nested_cls')

    global_ns.class_('nested_cls2')
    with pytest.raises(Exception):
        global_ns.class_('ns::nested_cls2')
    global_ns.class_('::ns::nested_cls2')

    global_ns.class_('nested_cls3')
    with pytest.raises(Exception):
        global_ns.class_('ns::nested_cls3')
    global_ns.class_('::ns::nested_cls3')



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""Defines :class:`scopedef_t` class"""

import timeit
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable

from . import algorithm
from . import templates
from . import declaration
from . import mdecl_wrapper
from . import byte_info
from . import elaborated_info
from . import runtime_errors
from .. import utils


class matcher(object):

    """Class-namespace, contains implementation of a few "find" algorithms"""

    @staticmethod
    def find(decl_matcher, decls, recursive=True):
        """
        Returns a list of declarations that match `decl_matcher` defined
        criteria or None

        :param decl_matcher: Python callable object, that takes one argument -
            reference to a declaration
        :param decls: the search scope, :class:declaration_t object or
            :class:declaration_t objects list t
        :param recursive: boolean, if True, the method will run `decl_matcher`
            on the internal declarations too
        """

        where = []
        if isinstance(decls, list):
            where.extend(decls)
        else:
            where.append(decls)
        if recursive:
            where = make_flatten(where)
        return list(filter(decl_matcher, where))

    @staticmethod
    def find_single(decl_matcher, decls, recursive=True):
        """
        Returns a reference to the declaration, that match `decl_matcher`
        defined criteria.

        if a unique declaration could not be found the method will return None.

        :param decl_matcher: Python callable object, that takes one argument -
            reference to a declaration
        :param decls: the search scope, :class:declaration_t object or
            :class:declaration_t objects list t
        :param recursive: boolean, if True, the method will run `decl_matcher`
            on the internal declarations too
        """
        answer = matcher.find(decl_matcher, decls, recursive)
        if len(answer) == 1:
            return answer[0]

    @staticmethod
    def get_single(decl_matcher, decls, recursive=True):
        """
        Returns a reference to declaration, that match `decl_matcher` defined
        criteria.

        If a unique declaration could not be found, an appropriate exception
        will be raised.

        :param decl_matcher: Python callable object, that takes one argument -
            reference to a declaration
        :param decls: the search scope, :class:declaration_t object or
            :class:declaration_t objects list t
        :param recursive: boolean, if True, the method will run `decl_matcher`
            on the internal declarations too
        """
        answer = matcher.find(decl_matcher, decls, recursive)
        if len(answer) == 1:
            return answer[0]
        elif not answer:
            raise runtime_errors.declaration_not_found_t(decl_matcher)
        else:
            raise runtime_errors.multiple_declarations_found_t(decl_matcher)


class scopedef_t(declaration.declaration_t):

    """
    Base class for :class:`namespace_t` and :class:`class_t` classes.

    This is the base class for all declaration classes that may have
    children nodes. The children can be accessed via the
    :attr:`scopedef_t.declarations` property.

    Also this class provides "get/select/find" interface. Using this class you
    can get instance or instances of internal declaration(s).

    You can find declaration(s) using next criteria:

        1. ``name`` - declaration name, could be full qualified name

        2. ``header_dir`` - directory, to which belongs file, that the
           declaration was declared in. ``header_dir`` should be absolute path.

        3. ``header_file`` - file that the declaration was declared in.

        4. ``function`` - user ( your ) custom criteria. The interesting
           thing is that this function will be joined with
           other arguments (criteria).

        5. ``recursive`` - the search declaration range, if True will be search
           in internal declarations too.

    Every ""query"" API, takes name or function as the first argument.

    .. code-block:: python

        global_namespace.member_function("do_something)

    the statement returns reference to member function named "do_something".
    If there the function doesn't exist or more than one function exists,
    an exception is raised.

    If you want to query for many declarations, use other function(s):

    .. code-block:: python

        do_something = global_namespace.member_functions("do_something")

    the statement returns :class:`mdecl_wrapper_t` instance. That object will
    save you writing `for` loops. For more information see
    :class:`the class <mdecl_wrapper_t>` documentation.
    """

    RECURSIVE_DEFAULT = True
    ALLOW_EMPTY_MDECL_WRAPPER = False

    # this class variable is used to prevent recursive imports
    _impl_matchers = {}
    # this class variable is used to prevent recursive imports
    _impl_decl_types = {}
    # this class variable is used to prevent recursive imports
    _impl_all_decl_types = []

    def __init__(self, name=''):
        declaration.declaration_t.__init__(self, name)

        self._optimized = False
        self._type2decls = {}
        self._type2name2decls = {}
        self._type2decls_nr = {}
        self._type2name2decls_nr = {}
        self._all_decls = None
        self._all_decls_not_recursive = []

    @property
    def _logger(self):
        """reference to :attr:`pygccxml.utils.loggers.queries_engine` logger"""
        return utils.loggers.queries_engine

    def _get__cmp__scope_items(self):
        """implementation details"""
        raise NotImplementedError()

    def _get__cmp__items(self):
        """implementation details"""
        items = []
        if self._optimized:
            # in this case we don't need to build class internal declarations
            # list
            items.append(self._all_decls_not_recursive.sort())
        else:
            items.append(self.declarations.sort())
        items.extend(self._get__cmp__scope_items())
        return items

    def __eq__(self, other):
        if not declaration.declaration_t.__eq__(self, other):
            return False
        return self.declarations[:].sort() == other.declarations[:].sort()

    __hash__ = declaration.declaration_t.__hash__

    def _get_declarations_impl(self):
        raise NotImplementedError()

    @property
    def declarations(self):
        """
        List of children declarations.

        Returns:
            List[declarations.declaration_t]
        """
        if self._optimized:
            return self._all_decls_not_recursive

        return self._get_declarations_impl()

    @declarations.setter
    def declarations(self, declarations):
        """
        Set list of all declarations defined in the namespace.

        Args:
            List[declarations.declaration_t]: list of declarations

        Not implemented.

        """
        raise NotImplementedError()

    def remove_declaration(self, decl):
        raise NotImplementedError()

    @staticmethod
    def __decl_types(decl):
        """implementation details"""
        types = []
        bases = list(decl.__class__.__bases__)
        if 'pygccxml' in decl.__class__.__module__:
            types.append(decl.__class__)
        while bases:
            base = bases.pop()
            if base is declaration.declaration_t:
                continue
            if base is byte_info.byte_info:
                continue
            if base is elaborated_info.elaborated_info:
                continue
            if 'pygccxml' not in base.__module__:
                continue
            types.append(base)
            bases.extend(base.__bases__)
        return types

    def clear_optimizer(self):
        """Cleans query optimizer state"""
        self._optimized = False
        self._type2decls = {}
        self._type2name2decls = {}
        self._type2decls_nr = {}
        self._type2name2decls_nr = {}
        self._all_decls = None
        self._all_decls_not_recursive = None

        for decl in self.declarations:
            if isinstance(decl, scopedef_t):
                decl.clear_optimizer()

    def init_optimizer(self):
        """
        Initializes query optimizer state.

        There are 4 internals hash tables:
            1. from type to declarations
            2. from type to declarations for non-recursive queries
            3. from type to name to declarations
            4. from type to name to declarations for non-recursive queries

        Almost every query includes declaration type information. Also very
        common query is to search some declaration(s) by name or full name.
        Those hash tables allows to search declaration very quick.
        """
        if self.name == '::':
            self._logger.debug(
                "preparing data structures for query optimizer - started")
        start_time = timeit.default_timer()

        self.clear_optimizer()

        for dtype in scopedef_t._impl_all_decl_types:
            self._type2decls[dtype] = []
            self._type2decls_nr[dtype] = []
            self._type2name2decls[dtype] = {}
            self._type2name2decls_nr[dtype] = {}

        self._all_decls_not_recursive = self.declarations
        self._all_decls = make_flatten(
            self._all_decls_not_recursive)
        for decl in self._all_decls:
            types = self.__decl_types(decl)
            for type_ in types:
                self._type2decls[type_].append(decl)
                name2decls = self._type2name2decls[type_]
                if decl.name not in name2decls:
                    name2decls[decl.name] = []
                name2decls[decl.name].append(decl)
                if self is decl.parent:
                    self._type2decls_nr[type_].append(decl)
                    name2decls_nr = self._type2name2decls_nr[type_]
                    if decl.name not in name2decls_nr:
                        name2decls_nr[decl.name] = []
                    name2decls_nr[decl.name].append(decl)

        for decl in self._all_decls_not_recursive:
            if isinstance(decl, scopedef_t):
                decl.init_optimizer()
        if self.name == '::':
            self._logger.debug((
                "preparing data structures for query optimizer - " +
                "done( %f seconds ). "), (timeit.default_timer() - start_time))
        self._optimized = True

    @staticmethod
    def _build_operator_function(name, function):
        if isinstance(name, Callable):
            return name

        return function

    @staticmethod
    def _build_operator_name(name, function, symbol):
        """implementation details"""
        def add_operator(sym):
            if 'new' in sym or 'delete' in sym:
                return 'operator ' + sym
            return 'operator' + sym
        if isinstance(name, Callable) and None is function:
            name = None
        if name:
            if 'operator' not in name:
                name = add_operator(name)
            return name
        elif symbol:
            return add_operator(symbol)
        return name  # both name and symbol are None

    def _on_rename(self):
        for decl in self.decls(allow_empty=True):
            decl.cache.reset_name_based()
        # I am not sure whether to introduce this or not?
        # It could be very time consuming operation + it changes optimize query
        # data structures.
        # if self.parent:
        #    if self.parent._optimized:
        #        self.parent.init_optimizer()

    @staticmethod
    def __normalize_args(**keywds):
        """implementation details"""
        if isinstance(keywds['name'], Callable) and \
                None is keywds['function']:
            keywds['function'] = keywds['name']
            keywds['name'] = None
        return keywds

    def __findout_recursive(self, **keywds):
        """implementation details"""
        if None is keywds['recursive']:
            return self.RECURSIVE_DEFAULT

        return keywds['recursive']

    def __findout_allow_empty(self, **keywds):
        """implementation details"""
        if None is keywds['allow_empty']:
            return self.ALLOW_EMPTY_MDECL_WRAPPER

        return keywds['allow_empty']

    @staticmethod
    def __findout_decl_type(match_class, **keywds):
        """implementation details"""
        if 'decl_type' in keywds:
            return keywds['decl_type']

        matcher_args = keywds.copy()
        del matcher_args['function']
        del matcher_args['recursive']
        if 'allow_empty' in matcher_args:
            del matcher_args['allow_empty']

        decl_matcher = match_class(**matcher_args)
        if decl_matcher.decl_type:
            return decl_matcher.decl_type
        return None

    def __create_matcher(self, match_class, **keywds):
        """implementation details"""
        matcher_args = keywds.copy()
        del matcher_args['function']
        del matcher_args['recursive']
        if 'allow_empty' in matcher_args:
            del matcher_args['allow_empty']

        decl_matcher = decl_matcher = match_class(**matcher_args)
        if keywds['function']:
            self._logger.debug(
                'running query: %s and <user defined function>',
                str(decl_matcher))
            return lambda decl: decl_matcher(decl) and keywds['function'](decl)

        self._logger.debug('running query: %s', str(decl_matcher))
        return decl_matcher

    def __findout_range(self, name, decl_type, recursive):
        """implementation details"""
        if not self._optimized:
            self._logger.debug(
                'running non optimized query - optimization has not been done')
            decls = self.declarations
            if recursive:
                decls = make_flatten(self.declarations)
            if decl_type:
                decls = [d for d in decls if isinstance(d, decl_type)]
            return decls

        if name and templates.is_instantiation(name):
            # templates has tricky mode to compare them, so lets check the
            # whole range
            name = None

        if name and decl_type:
            impl_match = scopedef_t._impl_matchers[scopedef_t.decl](name=name)
            if impl_match.is_full_name():
                name = impl_match.decl_name_only
            if recursive:
                self._logger.debug(
                    'query has been optimized on type and name')
                return self._type2name2decls[decl_type].get(name, [])

            self._logger.debug(
                'non recursive query has been optimized on type and name')
            return self._type2name2decls_nr[decl_type].get(name, [])
        elif decl_type:
            if recursive:
                self._logger.debug('query has been optimized on type')
                return self._type2decls[decl_type]

            self._logger.debug(
                'non recursive query has been optimized on type')
            return self._type2decls_nr[decl_type]
        else:
            if recursive:
                self._logger.debug((
                    'query has not been optimized ( hint: query does not ' +
                    'contain type and/or name )'))
                return self._all_decls

            self._logger.debug((
                'non recursive query has not been optimized ( hint: ' +
                'query does not contain type and/or name )'))
            return self._all_decls_not_recursive

    def _find_single(self, match_class, **keywds):
        """implementation details"""
        self._logger.debug('find single query execution - started')
        start_time = timeit.default_timer()
        norm_keywds = self.__normalize_args(**keywds)
        decl_matcher = self.__create_matcher(match_class, **norm_keywds)
        dtype = self.__findout_decl_type(match_class, **norm_keywds)
        recursive_ = self.__findout_recursive(**norm_keywds)
        decls = self.__findout_range(norm_keywds['name'], dtype, recursive_)
        found = matcher.get_single(decl_matcher, decls, False)
        self._logger.debug(
            'find single query execution - done( %f seconds )',
            (timeit.default_timer() - start_time))
        return found

    def _find_multiple(self, match_class, **keywds):
        """implementation details"""
        self._logger.debug('find all query execution - started')
        start_time = timeit.default_timer()
        norm_keywds = self.__normalize_args(**keywds)
        decl_matcher = self.__create_matcher(match_class, **norm_keywds)
        dtype = self.__findout_decl_type(match_class, **norm_keywds)
        recursive_ = self.__findout_recursive(**norm_keywds)
        allow_empty = self.__findout_allow_empty(**norm_keywds)
        decls = self.__findout_range(norm_keywds['name'], dtype, recursive_)
        found = matcher.find(decl_matcher, decls, False)
        mfound = mdecl_wrapper.mdecl_wrapper_t(found)
        self._logger.debug('%d declaration(s) that match query', len(mfound))
        self._logger.debug(
            'find single query execution - done( %f seconds )',
            (timeit.default_timer() - start_time))
        if not mfound and not allow_empty:
            raise RuntimeError(
                "Multi declaration query returned 0 declarations.")
        return mfound

    def decl(
            self,
            name=None,
            function=None,
            decl_type=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to declaration, that is matched defined
        criteria"""
        return (
            self._find_single(
                self._impl_matchers[
                    scopedef_t.decl],
                name=name,
                function=function,
                decl_type=decl_type,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def decls(
            self,
            name=None,
            function=None,
            decl_type=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of declarations, that are matched defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[
                    scopedef_t.decl],
                name=name,
                function=function,
                decl_type=decl_type,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def class_(
            self,
            name=None,
            function=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to class declaration, that is matched defined
        criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.class_],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.class_],
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def classes(
            self,
            name=None,
            function=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of class declarations, that are matched defined
        criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.class_],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.class_],
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def variable(
            self,
            name=None,
            function=None,
            decl_type=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to variable declaration, that is matched defined
        criteria"""

        return (
            self._find_single(
                self._impl_matchers[
                    scopedef_t.variable],
                name=name,
                function=function,
                decl_type=decl_type,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def variables(
            self,
            name=None,
            function=None,
            decl_type=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):

        """returns a set of variable declarations, that are matched defined
        criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[
                    scopedef_t.variable],
                name=name,
                function=function,
                decl_type=decl_type,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def calldef(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to "calldef" declaration, that is matched defined
        criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.calldef],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.calldef],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def calldefs(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of :class:`calldef_t` declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.calldef],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.calldef],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def operator(
            self,
            name=None,
            function=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to operator declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.operator],
                name=self._build_operator_name(name,
                                               function,
                                               symbol),
                symbol=symbol,
                function=self._build_operator_function(name,
                                                       function),
                decl_type=self._impl_decl_types[scopedef_t.operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def operators(
            self,
            name=None,
            function=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of operator declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.operator],
                name=self._build_operator_name(name,
                                               function,
                                               symbol),
                symbol=symbol,
                function=self._build_operator_function(name,
                                                       function),
                decl_type=self._impl_decl_types[scopedef_t.operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def member_function(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to member declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.member_function],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.member_function],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def member_functions(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of member function declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.member_function],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.member_function],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def constructor(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to constructor declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.constructor],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.constructor],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def constructors(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of constructor declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.constructor],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.constructor],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def member_operator(
            self,
            name=None,
            function=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to member operator declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.member_operator],
                name=self._build_operator_name(name,
                                               function,
                                               symbol),
                symbol=symbol,
                function=self._build_operator_function(name,
                                                       function),
                decl_type=self._impl_decl_types[scopedef_t.member_operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def member_operators(
            self,
            name=None,
            function=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of member operator declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.member_operator],
                name=self._build_operator_name(name,
                                               function,
                                               symbol),
                symbol=symbol,
                function=self._build_operator_function(name,
                                                       function),
                decl_type=self._impl_decl_types[scopedef_t.member_operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def casting_operator(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to casting operator declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.casting_operator],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.casting_operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def casting_operators(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of casting operator declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.casting_operator],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.casting_operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def enumeration(
            self,
            name=None,
            function=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to enumeration declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.enumeration],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.enumeration],
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def enumerations(
            self,
            name=None,
            function=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of enumeration declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.enumeration],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.enumeration],
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def typedef(
            self,
            name=None,
            function=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """returns reference to typedef declaration, that is matched
        defined criteria"""
        return (
            self._find_single(
                self._impl_matchers[scopedef_t.typedef],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.typedef],
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def typedefs(
            self,
            name=None,
            function=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """returns a set of typedef declarations, that are matched
        defined criteria"""
        return (
            self._find_multiple(
                self._impl_matchers[scopedef_t.typedef],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[
                    scopedef_t.typedef],
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def __getitem__(self, name_or_function):
        """
        Allow simple name based find of declarations.  Internally just calls
        `decls` method.
        :param name_or_function:  Name of `decl` to lookup or finder function.
        """
        return self.decls(name_or_function)

    def i_depend_on_them(self, recursive=True):
        raise NotImplementedError()


def make_flatten(decl_or_decls):
    """
    Converts tree representation of declarations to flatten one.

    :param decl_or_decls: reference to list of declaration's or single
        declaration
    :type decl_or_decls: :class:`declaration_t` or [ :class:`declaration_t` ]
    :rtype: [ all internal declarations ]

    """

    def proceed_single(decl):
        answer = [decl]
        if not isinstance(decl, scopedef_t):
            return answer
        for elem in decl.declarations:
            if isinstance(elem, scopedef_t):
                answer.extend(proceed_single(elem))
            else:
                answer.append(elem)
        return answer

    decls = []
    if isinstance(decl_or_decls, list):
        decls.extend(decl_or_decls)
    else:
        decls.append(decl_or_decls)
    answer = []
    for decl in decls:
        answer.extend(proceed_single(decl))
    return answer


def find_all_declarations(
        declarations,
        decl_type=None,
        name=None,
        parent=None,
        recursive=True,
        fullname=None):
    """
    Returns a list of all declarations that match criteria, defined by
    developer.

    For more information about arguments see :class:`match_declaration_t`
    class.

    :rtype: [ matched declarations ]

    """

    if recursive:
        decls = make_flatten(declarations)
    else:
        decls = declarations

    return list(
        filter(
            algorithm.match_declaration_t(
                decl_type=decl_type,
                name=name,
                fullname=fullname,
                parent=parent),
            decls))


def find_declaration(
        declarations,
        decl_type=None,
        name=None,
        parent=None,
        recursive=True,
        fullname=None):
    """
    Returns single declaration that match criteria, defined by developer.
    If more the one declaration was found None will be returned.

    For more information about arguments see :class:`match_declaration_t`
    class.

    :rtype: matched declaration :class:`declaration_t` or None

    """

    decl = find_all_declarations(
        declarations,
        decl_type=decl_type,
        name=name,
        parent=parent,
        recursive=recursive,
        fullname=fullname)
    if len(decl) == 1:
        return decl[0]


def find_first_declaration(
        declarations,
        decl_type=None,
        name=None,
        parent=None,
        recursive=True,
        fullname=None):
    """
    Returns first declaration that match criteria, defined by developer.

    For more information about arguments see :class:`match_declaration_t`
    class.

    :rtype: matched declaration :class:`declaration_t` or None

    """

    decl_matcher = algorithm.match_declaration_t(
        decl_type=decl_type,
        name=name,
        fullname=fullname,
        parent=parent)
    if recursive:
        decls = make_flatten(declarations)
    else:
        decls = declarations
    for decl in decls:
        if decl_matcher(decl):
            return decl
    return None


def declaration_files(decl_or_decls):
    """
    Returns set of files

    Every declaration is declared in some file. This function returns set, that
    contains all file names of declarations.

    :param decl_or_decls: reference to list of declaration's or single
        declaration
    :type decl_or_decls: :class:`declaration_t` or [:class:`declaration_t`]
    :rtype: set(declaration file names)

    """

    files = set()
    decls = make_flatten(decl_or_decls)
    for decl in decls:
        if decl.location:
            files.add(decl.location.file_name)
    return files



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines types visitor class interface
"""


class type_visitor_t(object):

    """
    types visitor interface

    All functions within this class should be redefined in derived classes.
    """

    def __init__(self):
        object.__init__(self)

    def visit_void(self):
        raise NotImplementedError()

    def visit_char(self):
        raise NotImplementedError()

    def visit_unsigned_char(self):
        raise NotImplementedError()

    def visit_signed_char(self):
        raise NotImplementedError()

    def visit_wchar(self):
        raise NotImplementedError()

    def visit_short_int(self):
        raise NotImplementedError()

    def visit_short_unsigned_int(self):
        raise NotImplementedError()

    def visit_bool(self):
        raise NotImplementedError()

    def visit_int(self):
        raise NotImplementedError()

    def visit_unsigned_int(self):
        raise NotImplementedError()

    def visit_long_int(self):
        raise NotImplementedError()

    def visit_long_unsigned_int(self):
        raise NotImplementedError()

    def visit_long_long_int(self):
        raise NotImplementedError()

    def visit_long_long_unsigned_int(self):
        raise NotImplementedError()

    def visit_int128(self):
        raise NotImplementedError()

    def visit_uint128(self):
        raise NotImplementedError()

    def visit_float(self):
        raise NotImplementedError()

    def visit_double(self):
        raise NotImplementedError()

    def visit_long_double(self):
        raise NotImplementedError()

    def visit_complex_long_double(self):
        raise NotImplementedError()

    def visit_complex_double(self):
        raise NotImplementedError()

    def visit_complex_float(self):
        raise NotImplementedError()

    def visit_jbyte(self):
        raise NotImplementedError()

    def visit_jshort(self):
        raise NotImplementedError()

    def visit_jint(self):
        raise NotImplementedError()

    def visit_jlong(self):
        raise NotImplementedError()

    def visit_jfloat(self):
        raise NotImplementedError()

    def visit_jdouble(self):
        raise NotImplementedError()

    def visit_jchar(self):
        raise NotImplementedError()

    def visit_jboolean(self):
        raise NotImplementedError()

    def visit_volatile(self):
        raise NotImplementedError()

    def visit_const(self):
        raise NotImplementedError()

    def visit_pointer(self):
        raise NotImplementedError()

    def visit_reference(self):
        raise NotImplementedError()

    def visit_elaborated(self):
        raise NotImplementedError()

    def visit_array(self):
        raise NotImplementedError()

    def visit_free_function_type(self):
        raise NotImplementedError()

    def visit_member_function_type(self):
        raise NotImplementedError()

    def visit_member_variable_type(self):
        raise NotImplementedError()

    def visit_declarated(self):
        raise NotImplementedError()

    def visit_restrict(self):
        raise NotImplementedError()

    def visit_ellipsis(self):
        raise NotImplementedError()



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find the location of the xml generator (castxml or gccxml)
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

# Parse the c++ file
decls = parser.parse([filename], xml_generator_config)

global_namespace = declarations.get_global_namespace(decls)

ns_namespace = global_namespace.namespace("ns")

int_type = declarations.cpptypes.int_t()
double_type = declarations.cpptypes.double_t()

# Search for the function by name
criteria = declarations.calldef_matcher(name="func1")
func1 = declarations.matcher.get_single(criteria, ns_namespace)

# Search for functions which return an int
criteria = declarations.calldef_matcher(return_type="int")
func2 = declarations.matcher.get_single(criteria, ns_namespace)

# Search for functions which return an int, using the cpptypes class
criteria = declarations.calldef_matcher(return_type=int_type)
func3 = declarations.matcher.get_single(criteria, ns_namespace)

print(func1)
print(func2)
print(func3)

# This prints:
# int ns::func1(int a) [free function]
# int ns::func1(int a) [free function]
# int ns::func1(int a) [free function]

# Search for functions which return a double. Two functions will be found
criteria = declarations.calldef_matcher(return_type=double_type)
func4 = declarations.matcher.find(criteria, ns_namespace)

print(len(func4))
print(func4[0])
print(func4[1])

# This prints:
# 2
# double ns::func2(double a) [free function]
# double ns::func3(double a) [free function]

# Finally, look for variables by name and by type
criteria = declarations.variable_matcher(name="a")
var_a1 = declarations.matcher.find(criteria, ns_namespace)

criteria = declarations.variable_matcher(decl_type=int_type)
var_a2 = declarations.matcher.find(criteria, ns_namespace)

print(var_a1[0])
print(var_a2[0])
print(var_a2[1])

# This prints:
# ns::a [variable]
# ns::a [variable]
# ns::b [variable]



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Defines :class:`pygccxml.declarations.declaration_t` class - all declarations
base class.

"""

import warnings

from . import declaration_utils
from . import algorithms_cache
from . import comment


class declaration_t(object):
    """
    Base class for all classes that represent a C++ declaration.

    """

    def __init__(
            self,
            name='',
            location=None,
            is_artificial=False,
            mangled=None,
            attributes=None):
        self._name = name
        self._location = location
        self._is_artificial = is_artificial
        self._mangled = mangled
        self._attributes = attributes
        self._parent = None
        self._cache = algorithms_cache.declaration_algs_cache_t()
        self._partial_name = None
        self._decorated_name = None
        self._comment = comment.comment_t()
        self._deprecation = None

    def __str__(self):
        """
        Default __str__ method.

        This version just returns the decl_string and the class.
        Derived classes may override this method to provide more detailed
        information.

        A __str__ method for a declaration should always provide enough
        information so that it uniquely identifies the declaration and
        the user is able to find the declaration in his source code.

        """

        name = self.decl_string
        if name[:2] == "::":
            name = name[2:]
        # Append the declaration class
        cls = self.__class__.__name__
        if cls[-2:] == "_t":
            cls = cls[:-2]
        cls = cls.replace('_', ' ')

        return "%s [%s]" % (name, cls)

    def _get__cmp__items(self):
        """
        Implementation detail.

        """
        # Every derived class should implement this method. This method should
        # return a list of items, that should be compared.

        print(
            '_get__cmp__items not implemented for class ',
            self.__class__.__name__)

        raise NotImplementedError()

    def _get__cmp__data(self):
        """
        Implementation detail.

        """
        if self._cache.cmp_data is None:
            cmp_data = [
                declaration_utils.declaration_path(self.parent),
                self.name,
                self.location]
            cmp_data.extend(self._get__cmp__items())
            self._cache.cmp_data = cmp_data

        return self._cache.cmp_data

    def __eq__(self, other):
        """
        This function will return true, if both declarations refers to the same
        object.
        This function could be implemented in terms of _get__cmp__data, but in
        this case it will downgrade performance. self.mangled property is not
        compared, because it could be changed from one compilation time to an
        other.

        """

        if not isinstance(other, self.__class__):
            return False
        return self.name == other.name \
            and self.location == other.location \
            and declaration_utils.declaration_path(self.parent) \
            == declaration_utils.declaration_path(other.parent)

    def __hash__(self):
        return (hash(self.__class__) ^
                hash(self.name) ^
                hash(self.location))

    def __ne__(self, other):
        """
        Return not self.__eq__( other ).

        """

        return not self.__eq__(other)

    def __lt__(self, other):
        """
        .. code-block:: python

           if not isinstance( other, self.__class__ ):
               return self.__class__.__name__ < other.__class__.__name__
           return self._get__cmp__data() < other._get__cmp__data()

        """

        if not isinstance(other, self.__class__):
            return self.__class__.__name__ < other.__class__.__name__
        return self._get__cmp__data() < other._get__cmp__data()

    def _get_name_impl(self):
        return self._name

    def _on_rename(self):
        """
        Placeholder method, is redefined in child class.

        """

        pass

    @property
    def name(self):
        """
        Declaration name
           @type: str

        """

        return self._get_name_impl()

    @name.setter
    def name(self, new_name):
        previous_name = self._name
        self._name = new_name
        self._partial_name = None
        self.cache.reset_name_based()
        if previous_name:
            # There was a reset of the name
            self._on_rename()

    def _get_partial_name_impl(self):
        return self.name

    @property
    def partial_name(self):
        """
        Declaration name, without template default arguments.

        Right now std containers is the only classes that support this
        functionality.

        """

        if None is self._partial_name:
            self._partial_name = self._get_partial_name_impl()

        return self._partial_name

    @property
    def parent(self):
        """
        Reference to parent declaration.

           @type: declaration_t

        """

        return self._parent

    @parent.setter
    def parent(self, new_parent):
        if new_parent:
            assert isinstance(new_parent, declaration_t)

        self._parent = new_parent

    @property
    def top_parent(self):
        """
        Reference to top parent declaration.

           @type: declaration_t

        """
        parent = self.parent
        while parent is not None:
            if parent.parent is None:
                return parent
            else:
                parent = parent.parent
        return self

    @property
    def location(self):
        """
        Location of the declaration within source file

           @type: :class:`location_t`

        """

        return self._location

    @location.setter
    def location(self, new_location):
        self._location = new_location

    @property
    def is_artificial(self):
        """
        Describes whether declaration is compiler generated or not

           @type: bool

        """

        return self._is_artificial

    @is_artificial.setter
    def is_artificial(self, new_artificial):
        self._is_artificial = bool(new_artificial)

    def get_mangled_name(self):
        return self._mangled

    @property
    def mangled(self):
        """
        Unique declaration name generated by the compiler.

        For GCCXML, you can get the mangled name for all the declarations.
        When using CastXML, calling mangled is only allowed on functions
        and variables. For other declarations it will raise an exception.

        :return: the mangled name
        :rtype: str

        """

        return self._mangled

    @mangled.setter
    def mangled(self, mangled):
        self._mangled = mangled

    @property
    def attributes(self):
        """
        GCCXML attributes, set using __attribute__((gccxml("...")))

           @type: str

        """

        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        self._attributes = attributes

    def create_decl_string(self, with_defaults=True):
        return declaration_utils.full_name(self, with_defaults)

    @property
    def decl_string(self):
        """
        Declaration full name.

        """

        return self.create_decl_string()

    @property
    def partial_decl_string(self):
        """
        Declaration full name.

        """

        return self.create_decl_string(with_defaults=False)

    @property
    def cache(self):
        """
        Implementation detail.

        Reference to instance of :class:`algorithms_cache_t` class.

        """

        return self._cache

    def i_depend_on_them(self, recursive=True):
        """
        Return list of all types and declarations the declaration depends on

        """
        self._warn_deprecated()
        raise NotImplementedError()

    @staticmethod
    def _warn_deprecated():
        """
        Implementation detail.
        """
        warnings.warn(
            "The i_depend_on_them() method is deprecated.\n" +
            "Please use the declarations.get_dependencies_from_decl()" +
            "function instead.\n",
            DeprecationWarning)

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        self._comment = comment

    @property
    def deprecation(self):
        return self._deprecation

    @deprecation.setter
    def deprecation(self, deprecation):
        self._deprecation = deprecation



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = ['test_elaborated_types.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_is_elaborated_type(global_ns):
    """
    Test for the is_elaborated function
    """

    for specifier in ["class", "struct", "enum", "union"]:
        _test_impl_yes(specifier=specifier, global_ns=global_ns)
        _test_impl_no(specifier=specifier, global_ns=global_ns)
        _test_arg_impl(specifier=specifier, global_ns=global_ns)


def _test_impl_yes(specifier, global_ns):
    yes = global_ns.namespace(name="::elaborated_t::yes_" + specifier)
    for decl in yes.declarations:
        assert declarations.is_elaborated(decl.decl_type) is True


def _test_impl_no(specifier, global_ns):
    no = global_ns.namespace(name="::elaborated_t::no_" + specifier)
    for decl in no.declarations:
        assert declarations.is_elaborated(decl.decl_type) is False


def _test_arg_impl(specifier, global_ns):
    decls = global_ns.namespace(
        name="::elaborated_t::arguments_" + specifier)
    for decl in decls.declarations:
        # The first argument is not elaborated
        no = decl.arguments[0].decl_type
        # The second argument is always elaborated
        yes = decl.arguments[1].decl_type
        assert declarations.is_elaborated(yes) is True
        assert declarations.is_elaborated(no) is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import declarations


def __test_split_impl(decl_string, name, args):
    assert (name, args) == \
        declarations.call_invocation.split(decl_string)


def __test_split_recursive_impl(decl_string, control_seq):
    assert control_seq == \
        list(declarations.call_invocation.split_recursive(decl_string))


def __test_is_call_invocation_impl(decl_string):
    assert declarations.call_invocation.is_call_invocation(decl_string)


def test_split_on_vector():
    __test_is_call_invocation_impl("vector(int,std::allocator(int) )")

    __test_split_impl(
        "vector(int,std::allocator(int) )",
        "vector",
        ["int", "std::allocator(int)"])

    __test_split_recursive_impl(
        "vector(int,std::allocator(int) )",
        [("vector", ["int", "std::allocator(int)"]),
            ("std::allocator", ["int"])])


def test_split_on_string():
    __test_is_call_invocation_impl(
        "basic_string(char,std::char_traits(char),std::allocator(char) )")

    __test_split_impl(
        "basic_string(char,std::char_traits(char),std::allocator(char) )",
        "basic_string",
        ["char", "std::char_traits(char)", "std::allocator(char)"])


def test_split_on_map():
    __test_is_call_invocation_impl(
        "map(long int,std::vector(int, std::allocator(int) )," +
        "std::less(long int),std::allocator(std::pair" +
        "(const long int, std::vector(int, std::allocator(int) ) ) ) )")

    __test_split_impl(
        "map(long int,std::vector(int, std::allocator(int) )," +
        "std::less(long int),std::allocator(std::pair" +
        "(const long int, std::vector(int, std::allocator(int) ) ) ) )",
        "map",
        ["long int", "std::vector(int, std::allocator(int) )",
            "std::less(long int)",
            "std::allocator(std::pair(const long int," +
            " std::vector(int, std::allocator(int) ) ) )"])


def test_join_on_vector():
    assert "vector(int, std::allocator(int))" == \
        declarations.call_invocation.join(
            "vector", ("int", "std::allocator(int)"))


def test_find_args():
    temp = 'x()()'
    found = declarations.call_invocation.find_args(temp)
    assert (1, 2) == found
    found = declarations.call_invocation.find_args(temp, found[1] + 1)
    assert (3, 4) == found
    temp = 'x(int,int)(1,2)'
    found = declarations.call_invocation.find_args(temp)
    assert (1, 9) == found
    found = declarations.call_invocation.find_args(temp, found[1] + 1)
    assert (10, 14) == found


def test_bug_unmatched_brace():
    src = 'AlternativeName((&string("")), (&string("")), (&string("")))'
    __test_split_impl(
        src, 'AlternativeName', [
            '(&string(""))', '(&string(""))', '(&string(""))'])



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ['test_order.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_order(global_ns):
    """
    Test order of const, volatile, etc... in decl_string.

    The convention in pygccxml is that const and volatile qualifiers
    are placed on the right of their `base` type.

    """
    c1 = global_ns.variable("c1")
    c2 = global_ns.variable("c2")
    assert c1.decl_type.decl_string == "int const"
    assert c2.decl_type.decl_string == "int const"

    cptr1 = global_ns.variable("cptr1")
    cptr2 = global_ns.variable("cptr2")
    assert cptr1.decl_type.decl_string == "int const * const"
    assert cptr2.decl_type.decl_string == "int const * const"

    v1 = global_ns.variable("v1")
    v2 = global_ns.variable("v2")
    assert v1.decl_type.decl_string == "int volatile"
    assert v2.decl_type.decl_string == "int volatile"

    vptr1 = global_ns.variable("vptr1")
    vptr2 = global_ns.variable("vptr2")
    decl_string = "int volatile * volatile"
    assert vptr1.decl_type.decl_string == decl_string
    assert vptr2.decl_type.decl_string == decl_string

    cv1 = global_ns.variable("cv1")
    cv2 = global_ns.variable("cv2")
    cv3 = global_ns.variable("cv3")
    cv4 = global_ns.variable("cv4")
    assert cv1.decl_type.decl_string == "int const volatile"
    assert cv2.decl_type.decl_string == "int const volatile"
    assert cv3.decl_type.decl_string == "int const volatile"
    assert cv4.decl_type.decl_string == "int const volatile"

    cvptr1 = global_ns.variable("cvptr1")
    cvptr2 = global_ns.variable("cvptr2")
    cvptr3 = global_ns.variable("cvptr3")
    cvptr4 = global_ns.variable("cvptr4")
    decl_string = "int const volatile * const volatile"
    assert cvptr1.decl_type.decl_string == decl_string
    assert cvptr2.decl_type.decl_string == decl_string
    assert cvptr3.decl_type.decl_string == decl_string
    assert cvptr4.decl_type.decl_string == decl_string

    ac1 = global_ns.variable("ac1")
    ac2 = global_ns.variable("ac2")
    assert ac1.decl_type.decl_string, "int const [2]"
    assert ac2.decl_type.decl_string, "int const [2]"

    acptr1 = global_ns.variable("acptr1")
    acptr2 = global_ns.variable("acptr2")
    assert acptr1.decl_type.decl_string == "int const * const [2]"
    assert acptr2.decl_type.decl_string == "int const * const [2]"

    class_a = global_ns.variable("classA")
    assert class_a.decl_type.decl_string == "::A const"



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


def test_source_reader_enums():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    reader = parser.source_reader_t(config)
    decls = reader.read_file("unnamed_enums_bug1.hpp")
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    names = []
    enums = global_ns.enumerations()
    for enum in enums:
        names.extend(list(enum.get_name2value_dict().keys()))
    assert len(names) == 4
    assert 'x1' in names
    assert 'x2' in names
    assert 'y1' in names
    assert 'y2' in names


def test_project_reader_enums():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(["unnamed_enums_bug1.hpp"], config)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()

    names = []
    for enum in global_ns.enumerations():
        names.extend(list(enum.get_name2value_dict().keys()))
    assert len(names) == 4
    assert 'x1' in names
    assert 'x2' in names
    assert 'y1' in names
    assert 'y2' in names


def test_multiple_files_enums():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(
        [
            'unnamed_enums_bug1.hpp',
            'unnamed_enums_bug2.hpp',
            'unnamed_enums_bug1.hpp'
        ], config
    )
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    names = []
    enums = global_ns.enumerations()
    list(map(
        lambda enum: names.extend(list(enum.get_name2value_dict().keys())),
        enums))
    assert len(names) == 6
    assert 'x1' in names
    assert 'x2' in names
    assert 'y1' in names
    assert 'y2' in names
    assert 'z1' in names
    assert 'z2' in names



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


def test_typedefs_src_reader():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    header = 'typedefs_base.hpp'
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse([header], config)
    global_ns = declarations.find_declaration(
        decls,
        decl_type=declarations.namespace_t,
        name='::')
    global_ns.init_optimizer()

    item_cls = global_ns.class_(name='item_t')
    assert item_cls is not None
    assert len(item_cls.aliases) == 1
    assert item_cls.aliases[0].name == 'Item'


def test_typedefs_source_reader():
    COMPILATION_MODE = parser.COMPILATION_MODE.FILE_BY_FILE
    config = autoconfig.cxx_parsers_cfg.config.clone()

    decls = parser.parse(
        ['typedefs1.hpp', 'typedefs2.hpp'],
        config,
        COMPILATION_MODE
    )
    item_cls = declarations.find_declaration(
        decls,
        decl_type=declarations.class_t,
        name='item_t')
    assert item_cls is not None
    assert len(item_cls.aliases) == 3
    expected_aliases = {'Item', 'Item1', 'Item2'}
    real_aliases = set([typedef.name for typedef in item_cls.aliases])
    assert real_aliases == expected_aliases



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import sys
import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    'core_ns_join_1.hpp',
    'core_ns_join_2.hpp',
    'core_ns_join_3.hpp',
    'core_membership.hpp',
    'core_class_hierarchy.hpp',
    'core_types.hpp',
    'core_diamand_hierarchy_base.hpp',
    'core_diamand_hierarchy_derived1.hpp',
    'core_diamand_hierarchy_derived2.hpp',
    'core_diamand_hierarchy_final_derived.hpp',
    'core_overloads_1.hpp',
    'core_overloads_2.hpp',
    'typedefs_base.hpp',
]


@pytest.fixture
def decls():
    COMPILATION_MODE = parser.COMPILATION_MODE.FILE_BY_FILE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    return decls


def test_printer(decls):
    # Redirect sys.stdout to a class with a writer doing nothing
    # This greatly reduces the size of the test output and makes
    # test log files readable.
    # Note: flush needs to be defined; because if not this will
    # result in an AttributeError on call.
    class DontPrint(object):
        def write(*args):
            pass

        def flush(*args):
            pass
    sys.stdout = DontPrint()

    declarations.print_declarations(decls)


def test__str__(decls):
    decls = declarations.make_flatten(decls)
    for decl in decls:
        str(decl)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

code = \
    """
template <typename T> struct A {};
template <int N> struct A<const char[N]>
{ static int size(const char[N]) { return N - 1; } };
"""


def test_partial_template():
    """
    The purpose of this test was to check if changes to GCCXML
    would lead to changes in the outputted xml file (Meaning
    the bug was fixed).

    GCCXML wrongly outputted partial template specialization.
    CastXML does not have this bug. In this case we check if
    the template specialization can not be found; which is the
    expected/wanted behaviour.

    https://github.com/CastXML/CastXML/issues/20

    """

    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse_string(code, config)
    global_ns = declarations.get_global_namespace(decls)
    with pytest.raises(declarations.declaration_not_found_t):
        global_ns.class_('A<const char [N]>')



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES1 = [
    'core_types.hpp',
    'core_ns_join_1.hpp',
    'core_ns_join_2.hpp',
    'core_ns_join_3.hpp',
    'core_membership.hpp',
    'core_class_hierarchy.hpp',
    'core_diamand_hierarchy_base.hpp',
    'core_diamand_hierarchy_derived1.hpp',
    'core_diamand_hierarchy_derived2.hpp',
    'core_diamand_hierarchy_final_derived.hpp',
    'core_overloads_1.hpp',
    'core_overloads_2.hpp'
]

TEST_FILES2 = [
    'separate_compilation/data.h',
    'separate_compilation/base.h',
    'separate_compilation/derived.h'
]


def test_correctness():
    for src in TEST_FILES1:
        __test_correctness_impl(src)


def __test_correctness_impl(file_name):
    config = autoconfig.cxx_parsers_cfg.config.clone()
    prj_reader = parser.project_reader_t(config)
    prj_decls = prj_reader.read_files(
        [file_name] * 2,
        compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)
    src_reader = parser.source_reader_t(config)
    src_decls = src_reader.read_file(file_name)
    if src_decls != prj_decls:
        s = src_decls[0]
        p = prj_decls[0]
        bdir = autoconfig.build_directory
        with open(os.path.join(bdir, file_name + '.sr.txt'), 'w+') as sr:
            with open(
                    os.path.join(bdir, file_name + '.pr.txt'), 'w+') as pr:

                declarations.print_declarations(
                    s, writer=lambda x: sr.write(l + os.linesep))
                declarations.print_declarations(
                    p, writer=lambda x: pr.write(l + os.linesep))
        raise (
            f"There is a difference between declarations in file {file_name}."
            )


def test_separate_compilation():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    prj_reader = parser.project_reader_t(config)
    prj_decls = prj_reader.read_files(
        TEST_FILES2,
        compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)
    src_reader = parser.source_reader_t(config)
    src_decls = src_reader.read_file('separate_compilation/all.h')

    declarations.dump_declarations(
        src_decls,
        os.path.join(
            autoconfig.build_directory, 'separate_compilation.sr.txt'))

    declarations.dump_declarations(
        prj_decls,
        os.path.join(
            autoconfig.build_directory, 'separate_compilation.pr.txt'))

    assert src_decls == prj_decls



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest
import platform

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    'test_pattern_parser.hpp'
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    if platform.system() == "Darwin":
        config.cflags = "-std=c++11 -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    else:
        config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_template_split_std_vector(global_ns):
    """
    Demonstrate error in pattern parser, see #60

    """

    config = autoconfig.cxx_parsers_cfg.config.clone()
    if platform.system() == "Darwin":
        config.cflags = "-std=c++11 -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    else:
        config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config)

    for decl in declarations.make_flatten(decls):
        if "myClass" in decl.name:
            _ = decl.partial_name


def test_matcher(global_ns):
    """
    Run the matcher on all the templated classes.

    This exercises the whole pipeline even more.

    """

    criteria = declarations.declaration_matcher(name="myClass")
    _ = declarations.matcher.find(criteria, global_ns)


def test_split():
    """
    Test a bunch of tricky name/args splits. More combinations could be
    tested but this is already covering most of the cases.

    In test_template_split_std_vector we test for a specific case that
    was failing (in a real world scenario).
    Here we test more possible combinations to make sure the splitting
    method is robust enough.

    """

    p1 = "std::vector<char, std::allocator<char> >"
    p2 = "std::vector<int, std::allocator<int> >"
    args_list = [
        "const std::basic_string<char> &", "const int &", "const double &"]

    for arg in args_list:

        li = [p1]
        name, args = declarations.templates.split(
            "myClass0a<" + ", ".join(li) + ">")
        assert name == "myClass0a"
        assert args == li

        li = [p1, p2]
        name, args = declarations.templates.split(
            "myClass0b<" + ", ".join(li) + ">")
        assert name == "myClass0b"
        assert args == li

        li = [p1, p2, p2]
        name, args = declarations.templates.split(
            "myClass0c<" + ", ".join(li) + ">")
        assert name == "myClass0c"
        assert args == li

        li = [p1 + " (" + arg + ")"]
        name, args = declarations.templates.split(
            "myClass1<" + ", ".join(li) + ">")
        assert name == "myClass1"
        assert args == li

        li = [p1 + " (" + arg + ", " + arg + ")"]
        name, args = declarations.templates.split(
            "myClass2<" + ", ".join(li) + ">")
        assert name == "myClass2"
        assert args == li

        li = [p2 + " (" + arg + ", " + arg + ")"]
        name, args = declarations.templates.split(
            "myClass3<" + ", ".join(li) + ">")
        assert name == "myClass3"
        assert args == li

        li = [p1 + " (" + arg + ", " + arg + ", " + arg + ")"]
        name, args = declarations.templates.split(
            "myClass4<" + ", ".join(li) + ">")
        assert name == "myClass4"
        assert args == li

        li = [
            p1 + " (" + arg + ", " + arg + ", " + arg + ")",
            p1]
        name, args = declarations.templates.split(
            "myClass5<" + ", ".join(li) + ">")
        assert name == "myClass5"
        assert args == li

        li = [
            p1,
            p1 + " (" + arg + ", " + arg + ", " + arg + ")"]
        name, args = declarations.templates.split(
            "myClass6<" + ", ".join(li) + ">")
        assert name == "myClass6"
        assert args == li

        li = [
            p2 + " (" + arg + ")",
            p1,
            p1 + " (" + arg + ", " + arg + ", " + arg + ")"]
        name, args = declarations.templates.split(
            "myClass7<" + ", ".join(li) + ">")
        assert name == "myClass7"
        assert args == li

        li = [
            p1,
            p2 + " (" + arg + ")",
            p1 + " (" + arg + ", " + arg + ", " + arg + ")"]
        name, args = declarations.templates.split(
            "myClass8<" + ", ".join(li) + ">")
        assert name == "myClass8"
        assert args == li

        li = [
            p2 + " (" + arg + ")",
            p1 + " (" + arg + ", " + arg + ")",
            p1]
        name, args = declarations.templates.split(
            "myClass9<" + ", ".join(li) + ">")
        assert name == "myClass9"
        assert args == li

        li = [
            p2 + " (" + arg + ")",
            p1 + " (" + arg + ", " + arg + ", " + arg + ")",
            p1,
            p2]
        name, args = declarations.templates.split(
            "myClass10<" + ", ".join(li) + ">")
        assert name == "myClass10"
        assert args == li



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "declaration_string.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_create_decl_string(global_ns):
    """
    Test the create_decl_string method.

    """

    myfunc = global_ns.free_function("myfunc")

    decl = declarations.free_function_type_t.create_decl_string(
        myfunc.return_type, myfunc.argument_types)

    assert decl != "('int (*)( int,int )', 'int (*)( int,int )')"

    box = global_ns.class_("Box")
    myinternfunc = box.member_function("myinternfunc")
    decl = declarations.member_function_type_t.create_decl_string(
        myinternfunc.return_type,
        box.decl_string,
        myinternfunc.argument_types,
        myinternfunc.has_const)

    assert decl != "short int ( ::Box::* )(  ) "



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    'core_ns_join_1.hpp',
    'core_ns_join_2.hpp',
    'core_ns_join_3.hpp',
    'core_membership.hpp',
    'core_class_hierarchy.hpp',
    'core_types.hpp',
    'core_diamand_hierarchy_base.hpp',
    'core_diamand_hierarchy_derived1.hpp',
    'core_diamand_hierarchy_derived2.hpp',
    'core_diamand_hierarchy_final_derived.hpp',
    'core_overloads_1.hpp',
    'core_overloads_2.hpp',
    'typedefs_base.hpp',
]


def test_declaration_files():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    prj_reader = parser.project_reader_t(config)
    decls = prj_reader.read_files(
        TEST_FILES,
        compilation_mode=parser.COMPILATION_MODE.ALL_AT_ONCE)
    files = declarations.declaration_files(decls)
    result = set()
    for fn in files:
        result.add(os.path.split(fn)[1])
    assert set(TEST_FILES).issubset(result) is True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES1 = [
    "bit_fields.hpp",
]

TEST_FILES2 = [
    "vector_traits.hpp",
]


def test_bit_fields():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES1, config, COMPILATION_MODE)

    criteria = declarations.variable_matcher_t(
        name='x',
        decl_type='unsigned int')
    x = declarations.matcher.get_single(criteria, decls)

    comp_str = (
        '(decl type==variable_t) and (name==x) and ' +
        '(value type==unsigned int)')
    assert str(criteria) == comp_str

    criteria = declarations.variable_matcher_t(
        name='::bit_fields::fields_t::x',
        decl_type=declarations.unsigned_int_t(),
        header_dir=os.path.dirname(
            x.location.file_name),
        header_file=x.location.file_name)

    x = declarations.matcher.get_single(criteria, decls)
    assert x is not None
    assert 'public' == x.access_type


def test_no_defaults():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES2, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)

    global_ns.decls(lambda decl: 'vector<' in decl.name)
    global_ns.decl('vector<_0_>')
    global_ns.class_('vector<std::vector<int>>')
    global_ns.class_('vector<std::string>')
    global_ns.decl('vector<const int>')



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import platform

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ["test_map_gcc5.hpp"]


def test_map_gcc5():
    """
    The code in test_map_gcc5.hpp was breaking pygccxml.

    Test that case (gcc5 + castxml + c++11).

    See issue #45 and #55

    """

    config = autoconfig.cxx_parsers_cfg.config.clone()
    if platform.system() == "Darwin":
        config.cflags = "-std=c++11 -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    else:
        config.cflags = "-std=c++11"

    decls = parser.parse(TEST_FILES, config)
    global_ns = declarations.get_global_namespace(decls)

    # This calldef is defined with gcc > 4.9 (maybe earlier, not tested)
    # and -std=c++11. Calling create_decl_string is failing with gcc.
    # With clang the calldef does not exist so the matcher
    # will just return an empty list, letting the test pass.
    # See the test_argument_without_name.py for an equivalent test,
    # which is not depending on the presence of the _M_clone_node
    # method in the stl_tree.h file.
    criteria = declarations.calldef_matcher(name="_M_clone_node")
    free_funcs = declarations.matcher.find(criteria, global_ns)
    for free_funcs in free_funcs:
        free_funcs.create_decl_string(with_defaults=False)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest
import warnings

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "include_all.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_variable(global_ns):
    ns_vars = global_ns.namespace('::declarations::variables')
    static_var = ns_vars.variable('static_var')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    warnings.simplefilter("ignore", Warning)
    dependencies_old = static_var.i_depend_on_them()
    warnings.simplefilter("error", Warning)
    assert len(dependencies_old) == 1
    assert dependencies_old[0].declaration is static_var
    assert dependencies_old[0].depend_on_it.decl_string == 'int'

    dependencies_new = declarations.get_dependencies_from_decl(static_var)
    assert len(dependencies_new) == 1
    assert dependencies_new[0].declaration is static_var
    assert dependencies_new[0].depend_on_it.decl_string == 'int'

    m_mutable = ns_vars.variable('m_mutable')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    warnings.simplefilter("ignore", Warning)
    dependencies_old = m_mutable.i_depend_on_them()
    warnings.simplefilter("error", Warning)
    assert len(dependencies_old) == 1
    assert dependencies_old[0].declaration is m_mutable
    assert dependencies_old[0].depend_on_it.decl_string == 'int'

    dependencies_new = declarations.get_dependencies_from_decl(m_mutable)
    assert len(dependencies_new) == 1
    assert dependencies_new[0].declaration is m_mutable
    assert dependencies_new[0].depend_on_it.decl_string == 'int'


def test_class(global_ns):
    ns_vars = global_ns.namespace('::declarations::variables')

    cls = ns_vars.class_('struct_variables_t')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    warnings.simplefilter("ignore", Warning)
    dependencies_old = cls.i_depend_on_them()
    warnings.simplefilter("error", Warning)
    dependencies_old = [
        d for d in dependencies_old if not d.declaration.is_artificial]
    assert len(dependencies_old) == 1

    dependencies_new = declarations.get_dependencies_from_decl(cls)
    dependencies_new = [
        d for d in dependencies_new if not d.declaration.is_artificial]
    assert len(dependencies_new) == 1

    m_mutable = ns_vars.variable('m_mutable')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    dependencies_old = [
        dependency for dependency in dependencies_old if
        dependency.declaration is m_mutable]
    assert len(dependencies_old) == 1
    assert dependencies_old[0].depend_on_it.decl_string == 'int'
    assert dependencies_old[0].access_type == 'public'

    dependencies_new = [
        dependency for dependency in dependencies_new if
        dependency.declaration is m_mutable]
    assert len(dependencies_new) == 1
    assert dependencies_new[0].depend_on_it.decl_string == 'int'
    assert dependencies_new[0].access_type == 'public'

    ns_dh = global_ns.namespace('::core::diamand_hierarchy')
    fd_cls = ns_dh.class_('final_derived_t')
    derived1_cls = ns_dh.class_('derived1_t')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    warnings.simplefilter("ignore", Warning)
    dependencies_old = declarations.get_dependencies_from_decl(fd_cls)
    warnings.simplefilter("error", Warning)
    dependencies_old = [
        dependency for dependency in dependencies_old if
        dependency.depend_on_it is derived1_cls]
    assert len(dependencies_old) == 1
    assert dependencies_old[0].depend_on_it is derived1_cls
    assert dependencies_old[0].access_type == 'public'

    dependencies_new = declarations.get_dependencies_from_decl(fd_cls)
    dependencies_new = [
        dependency for dependency in dependencies_new if
        dependency.depend_on_it is derived1_cls]
    assert len(dependencies_new) == 1
    assert dependencies_new[0].depend_on_it is derived1_cls
    assert dependencies_new[0].access_type == 'public'


def test_calldefs(global_ns):
    ns = global_ns.namespace('::declarations::calldef')
    return_default_args = ns.calldef('return_default_args')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    warnings.simplefilter("ignore", Warning)
    dependencies_old = return_default_args.i_depend_on_them()
    warnings.simplefilter("error", Warning)
    assert len(dependencies_old) == 3
    used_types = [
        dependency.depend_on_it.decl_string
        for dependency in dependencies_old]
    assert used_types == ['int', 'int', 'bool']

    dependencies_new = declarations.get_dependencies_from_decl(
        return_default_args)
    assert len(dependencies_new) == 3
    used_types = [
        dependency.depend_on_it.decl_string
        for dependency in dependencies_new]
    assert used_types == ['int', 'int', 'bool']


def test_coverage(global_ns):
    declarations.get_dependencies_from_decl(global_ns)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ["non_copyable_classes.hpp"]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test(global_ns):
    """
    Search for classes which can not be copied.

    See bug #13

    1) non copyable class
    2) non copyable const variable (fundamental type)
    3) non copyable const variable (class type)
    4) non copyable const variable (array type)
    5) non copyable const variable (class type)

    """

    main_foo_1 = global_ns.class_('MainFoo1')
    assert declarations.is_noncopyable(main_foo_1) is True

    main_foo_2 = global_ns.class_('MainFoo2')
    assert declarations.is_noncopyable(main_foo_2) is True

    main_foo_3 = global_ns.class_('MainFoo3')
    assert declarations.is_noncopyable(main_foo_3) is True

    main_foo_4 = global_ns.class_('MainFoo4')
    assert declarations.is_noncopyable(main_foo_4) is True

    main_foo_5 = global_ns.class_('MainFoo5')
    assert declarations.is_noncopyable(main_foo_5) is True

    main_foo_6 = global_ns.class_('MainFoo6')
    assert declarations.is_noncopyable(main_foo_6) is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest


class Helpers:
    @staticmethod
    def _test_type_composition(type_, expected_compound, expected_base):
        assert isinstance(type_, expected_compound)
        assert isinstance(type_.base, expected_base)

    @staticmethod
    def _test_calldef_args(calldef, expected_args):
        assert len(calldef.arguments) == len(expected_args)

        for i, expected_arg in enumerate(expected_args):
            arg = calldef.arguments[i]
            assert arg == expected_arg

    @staticmethod
    def _test_calldef_return_type(calldef, expected_type):
        assert isinstance(calldef.return_type, expected_type)

    @staticmethod
    def _test_calldef_exceptions(global_ns, calldef, exceptions):
        # exceptions is list of classes names
        exception_decls = []
        for name in exceptions:
            exception_decl = global_ns.class_(name)
            assert exception_decl is not None
            exception_decls.append(exception_decl)
        exception_decls.sort()
        assert len(calldef.exceptions) == len(exception_decls)
        exceptions_indeed = sorted(calldef.exceptions[:])
        assert exception_decls == exceptions_indeed


@pytest.fixture
def helpers():
    return Helpers



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)
ns = global_namespace.namespace("ns")

# We have just one declaration in ns, which is our Test class:
classTest = ns.declarations[0]
print(classTest.name, type(classTest))
# > 'Test', <class 'pygccxml.declarations.class_declaration.class_t'

# Loop over the two constructors:
for constructor in classTest.constructors():
    print(str(constructor), constructor.is_artificial)
# > ns::Test::Test() [constructor], False
# > ns::Test::Test(ns::Test const & arg0) [constructor], True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import os.path

from . import autoconfig

from pygccxml.parser.config import xml_generator_configuration_t
from pygccxml.parser import declarations_cache


def test_file_signature():
    file1 = os.path.join(autoconfig.data_directory, 'decl_cache_file1.txt')
    file1_dup = os.path.join(
        autoconfig.data_directory,
        'decl_cache_file1_duplicate.txt')
    file2 = os.path.join(autoconfig.data_directory, 'decl_cache_file2.txt')
    sig1 = declarations_cache.file_signature(file1)
    sig1_dup = declarations_cache.file_signature(file1_dup)
    sig2 = declarations_cache.file_signature(file2)
    assert sig1 == sig1_dup
    assert sig1 != sig2


def test_config_signature():
    diff_cfg_list = build_differing_cfg_list()
    def_cfg = diff_cfg_list[0]
    def_sig = declarations_cache.configuration_signature(def_cfg)

    # Test changes that should cause sig changes
    for cfg in diff_cfg_list[1:]:
        assert declarations_cache.configuration_signature(cfg) != def_sig

    # Test changes that should not cause sig changes
    no_changes = def_cfg.clone()
    assert declarations_cache.configuration_signature(no_changes) == def_sig

    ignore_changed = def_cfg.clone()
    ignore_changed.ignore_gccxml_output = True
    assert (
        declarations_cache.configuration_signature(ignore_changed) == def_sig
        )


def test_cache_interface():
    cache_file = os.path.join(
        autoconfig.build_directory,
        'decl_cache_test.test_cache_read.cache')
    file1 = os.path.join(autoconfig.data_directory, 'decl_cache_file1.txt')
    file1_dup = os.path.join(
        autoconfig.data_directory,
        'decl_cache_file1_duplicate.txt')
    file2 = os.path.join(autoconfig.data_directory, 'decl_cache_file2.txt')
    diff_cfg_list = build_differing_cfg_list()
    def_cfg = diff_cfg_list[0]

    if os.path.exists(cache_file):
        os.remove(cache_file)

    cache = declarations_cache.file_cache_t(cache_file)
    assert len(cache._file_cache_t__cache) == 0

    # test creating new entries for differing files
    cache.update(file1, def_cfg, 1, [])
    assert len(cache._file_cache_t__cache) == 1
    cache.update(file1_dup, def_cfg, 2, [])
    assert len(cache._file_cache_t__cache) == 1
    cache.update(file2, def_cfg, 3, [])
    assert len(cache._file_cache_t__cache) == 2

    assert cache.cached_value(file1, def_cfg) == 2
    assert cache.cached_value(file2, def_cfg) == 3

    # Test reading again
    cache.flush()
    cache = declarations_cache.file_cache_t(cache_file)
    assert len(cache._file_cache_t__cache) == 2
    assert cache.cached_value(file1, def_cfg) == 2
    assert cache.cached_value(file2, def_cfg) == 3

    # Test flushing doesn't happen if we don't touch the cache
    cache = declarations_cache.file_cache_t(cache_file)
    assert cache.cached_value(file1, def_cfg) == 2  # Read from cache
    cache.flush()    # should not actually flush
    cache = declarations_cache.file_cache_t(cache_file)
    assert len(cache._file_cache_t__cache) == 2

    # Test flush culling
    cache = declarations_cache.file_cache_t(cache_file)
    cache.update(file1_dup, def_cfg, 4, [])    # Modify cache
    cache.flush()    # should cull off one entry
    cache = declarations_cache.file_cache_t(cache_file)
    assert len(cache._file_cache_t__cache) == 1


def build_differing_cfg_list():
    """ Return a list of configurations that all differ. """
    cfg_list = []
    def_cfg = xml_generator_configuration_t(
        "xml_generator_path",
        '.', ['tmp'], ['sym'], ['unsym'], None, False, "")
    cfg_list.append(def_cfg)

    # Test changes that should cause sig changes
    gccxml_changed = def_cfg.clone()
    gccxml_changed.xml_generator_path = "other_path"
    cfg_list.append(gccxml_changed)

    wd_changed = def_cfg.clone()
    wd_changed.working_directory = "other_dir"
    cfg_list.append(wd_changed)

    # inc_changed = def_cfg.clone()
    # inc_changed.include_paths = ["/var/tmp"]
    # assert configuration_signature(inc_changed) != def_sig)
    inc_changed = xml_generator_configuration_t(
        "xml_generator_path", '.', ['/var/tmp'], ['sym'], ['unsym'],
        None, False, "")
    cfg_list.append(inc_changed)

    # def_changed = def_cfg.clone()
    # def_changed.define_symbols = ["symbol"]
    # assert configuration_signature(def_changed) != def_sig)
    def_changed = xml_generator_configuration_t(
        "xml_generator_path", '.', ['/var/tmp'], ['new-sym'], ['unsym'],
        None, False, "")
    cfg_list.append(def_changed)

    # undef_changed = def_cfg.clone()
    # undef_changed.undefine_symbols = ["symbol"]
    # assert configuration_signature(undef_changed) != def_sig)
    undef_changed = xml_generator_configuration_t(
        "xml_generator_path", '.', ['/var/tmp'], ['sym'], ['new-unsym'],
        None, False, "")
    cfg_list.append(undef_changed)

    cflags_changed = def_cfg.clone()
    cflags_changed.cflags = "new flags"
    cfg_list.append(cflags_changed)

    return cfg_list



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

import platform

from . import autoconfig

from pygccxml import parser


def test_cpp_standards():
    """
    Test different compilation standards by setting cflags.

    """

    config = autoconfig.cxx_parsers_cfg.config.clone()

    cflags_common = ""

    if platform.system() == "Darwin":
        cflags_common = " -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01

    parser.parse(["cpp_standards.hpp"], config)

    if platform.system() != 'Windows':
        config.cflags = "-std=c++98" + cflags_common
        parser.parse(["cpp_standards.hpp"], config)

        config.cflags = "-std=c++03" + cflags_common
        parser.parse(["cpp_standards.hpp"], config)

    config.cflags = "-std=c++11" + cflags_common

    parser.parse(["cpp_standards.hpp"], config)

    config.cflags = "-std=c++14" + cflags_common
    parser.parse(["cpp_standards.hpp"], config)

    config.cflags = "-std=c++1z" + cflags_common
    parser.parse(["cpp_standards.hpp"], config)

    # Pass down a flag that does not exist.
    # This should raise an exception.
    config.cflags = "-std=c++00" + cflags_common
    with pytest.raises(RuntimeError):
        parser.parse(["cpp_standards.hpp"], config)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "bit_fields.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_bit_fields(global_ns):
    bf_x = global_ns.variable('x')
    assert bf_x.bits == 1

    bf_y = global_ns.variable('y')
    assert bf_y.bits == 7

    mv_z = global_ns.variable('z')
    assert mv_z.bits is None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from pygccxml import parser
from pygccxml import utils


def test_config():
    """Test config setup with wrong xml generator setups."""

    # Some code to parse for the example
    code = "int a;"

    # Find the location of the xml generator (castxml)
    generator_path, name = utils.find_xml_generator()

    # No xml generator path
    config = parser.xml_generator_configuration_t(xml_generator=name)
    with pytest.raises(RuntimeError):
        parser.parse_string(code, config)

    # Invalid path
    config = parser.xml_generator_configuration_t(
        xml_generator_path="wrong/path",
        xml_generator=name)
    with pytest.raises(RuntimeError):
        parser.parse_string(code, config)

    # None path
    config = parser.xml_generator_configuration_t(
        xml_generator_path=None,
        xml_generator=name)
    with pytest.raises(RuntimeError):
        parser.parse_string(code, config)

    # No name
    config = parser.xml_generator_configuration_t(
        xml_generator_path=generator_path)
    with pytest.raises(RuntimeError):
        parser.parse_string(code, config)

    # Random name
    config = parser.xml_generator_configuration_t(
        xml_generator_path=generator_path,
        xml_generator="not_a_generator")
    with pytest.raises(RuntimeError):
        parser.parse_string(code, config)

    # None name
    config = parser.xml_generator_configuration_t(
        xml_generator_path=generator_path,
        xml_generator=None)
    with pytest.raises(RuntimeError):
        parser.parse_string(code, config)



# Copyright 2014-2020 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "test_comments.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def _check_comment_content(list, comment_decl):
    if comment_decl.text:
        assert list == comment_decl.text
    else:
        print("No text in comment to check")


def test_comments(global_ns):
    """
    Check the comment parsing
    """
    tnamespace = global_ns.namespace("comment")

    assert "comment" in dir(tnamespace)
    _check_comment_content(
        [
            "//! Namespace Comment",
            "//! Across multiple lines"
        ],
        tnamespace.comment
    )

    tenumeration = tnamespace.enumeration("com_enum")
    assert "comment" in dir(tenumeration)
    _check_comment_content(
        ['/// Outside Class enum comment'],
        tenumeration.comment
    )

    tclass = tnamespace.class_("test")
    assert "comment" in dir(tclass)
    _check_comment_content(
        ["/** class comment */"],
        tclass.comment
    )

    tcls_enumeration = tclass.enumeration("test_enum")
    assert "comment" in dir(tcls_enumeration)
    _check_comment_content(
        ['/// inside class enum comment'],
        tcls_enumeration.comment
    )

    tmethod = tclass.member_functions()[0]

    assert "comment" in dir(tmethod)
    _check_comment_content(
        ["/// cxx comment", "/// with multiple lines"],
        tmethod.comment
    )

    tconstructor = tclass.constructors()[0]

    assert "comment" in dir(tconstructor)
    _check_comment_content(
        ["/** doc comment */"],
        tconstructor.comment
    )

    for indx, cmt in enumerate(
                [
                    '//! mutable field comment',
                    "/// bit field comment"
                ]
            ):
        tvariable = tclass.variables()[indx]
        assert "comment" in dir(tvariable)
        _check_comment_content([cmt], tvariable.comment)



# Copyright 2021 Insight Software Consortium.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "test_deprecation.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def _check_text_content(desired_text, deprecation_string):
    assert desired_text == deprecation_string


def test_comment_deprecation(global_ns):
    """
    Check the comment parsing
    """

    tnamespace = global_ns.namespace("deprecation")

    tenumeration = tnamespace.enumeration("com_enum")
    assert "deprecation" in dir(tenumeration)
    _check_text_content(
        'Enumeration is Deprecated',
        tenumeration.deprecation)

    tclass = tnamespace.class_("test")
    assert "deprecation" in dir(tclass)
    _check_text_content(
        "Test class Deprecated",
        tclass.deprecation)

    tmethod = tclass.member_functions()[0]
    tmethod_dep = tclass.member_functions()[1]

    assert "deprecation", dir(tmethod)
    assert tmethod.deprecation is None
    _check_text_content(
        "Function is deprecated",
        tmethod_dep.deprecation)

    tconstructor = tclass.constructors()[0]
    tconstructor_dep = tclass.constructors()[1]

    assert tconstructor.deprecation is None
    assert "deprecation" in dir(tconstructor_dep)
    _check_text_content(
        "One arg constructor is Deprecated",
        tconstructor_dep.deprecation)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import platform

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    'declarations_calldef.hpp'
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    if platform.system() == "Darwin":
        config.cflags = "-Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_compound_argument_type(global_ns):
    do_smth = global_ns.calldefs('do_smth')
    assert do_smth is not None
    do_smth.function_type()



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    'plain_c.c'
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_plain_c(global_ns):
    global_ns.free_function('hello_sum')
    global_ns.free_function('hello_print')
    f = global_ns.free_function('do_smth')
    for arg in f.arguments:
        assert arg.decl_type.decl_string is not None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)
ns = global_namespace.namespace("ns")

a = ns.variables()[0]

print("My name is: " + a.name)
# > My name is: a

print("My type is: " + str(a.decl_type))
# > My type is: int const

# If we print real python type:
print("My type is : " + str(type(a.decl_type)))
# > My type is: <class 'pygccxml.declarations.cpptypes.const_t'>

# Types are nested in pygccxml. This means that you will get information
# about the first type only. You can access the "base" type by removing
# the const part:
print("My base type is: " + str(type(declarations.remove_const(a.decl_type))))
# > My base type is: <class 'pygccxml.declarations.cpptypes.int_t'>

# You use the is_const function to check for a type:
print("Is 'a' a const ?: " + str(declarations.is_const(a.decl_type)))
# > Is 'a' a const ?: True

# A more complex example with variable b:
b = ns.variables()[1]
print("My type is: " + str(type(b.decl_type)))
# > My type is: <class 'pygccxml.declarations.cpptypes.pointer_t'>
print("My type is: " + str(type(
    declarations.remove_const(
        declarations.remove_volatile(
            declarations.remove_pointer(b.decl_type))))))
# > My type is: <class 'pygccxml.declarations.cpptypes.int_t'>

# The declarations module contains much more methods allowing you to
# navigate the nested types list.



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser

TEST_FILE = os.path.join(autoconfig.data_directory, 'core_cache.hpp')
cache_file = os.path.join(
            autoconfig.data_directory,
            'pygccxml.cache')


def reset_cache():
    if os.path.exists(cache_file) and os.path.isfile(cache_file):
        os.remove(cache_file)


def touch():
    # Need to change file.
    with open(TEST_FILE, "a+") as header:
        header.write("//touch")


def test_update_cache():
    reset_cache()
    config = autoconfig.cxx_parsers_cfg.config.clone()

    # Save the content of the header file for later
    with open(TEST_FILE, "r") as old_header:
        content = old_header.read()

    declarations = parser.parse([TEST_FILE], config)
    cache = parser.file_cache_t(cache_file)
    cache.update(
        source_file=TEST_FILE,
        configuration=config,
        declarations=declarations,
        included_files=[])
    assert declarations == cache.cached_value(
            TEST_FILE,
            config)
    touch()
    assert cache.cached_value(TEST_FILE, config) is None

    # We wrote a //touch in the header file. Just replace the file with the
    # original content. The touched file would be sometimes commited by
    # error as it was modified.
    with open(TEST_FILE, "w") as new_header:
        new_header.write(content)


def test_cache_from_file():
    reset_cache()
    config = autoconfig.cxx_parsers_cfg.config.clone()
    declarations = parser.parse([TEST_FILE], config)
    cache = parser.file_cache_t(cache_file)
    cache.update(
        source_file=TEST_FILE,
        configuration=config,
        declarations=declarations,
        included_files=[])
    assert declarations == cache.cached_value(
            TEST_FILE,
            config)
    cache.flush()
    cache = parser.file_cache_t(cache_file)
    assert declarations == cache.cached_value(
            TEST_FILE,
            config)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import warnings

from pygccxml import utils


def test_contains_parent_dir():
    path = os.path.normpath("/mypath/folder1/folder2/folder3")
    dirs = [
        os.path.normpath("/mypath/folder1/folder2/"),
        os.path.normpath("/mypath3/folder1/folder2/folder3"),
        os.path.normpath("home"),
        os.path.normpath("/test/test1/mypath")]

    assert utils.utils.contains_parent_dir(path, dirs) is True

    dirs = [os.path.normpath("/home"), os.path.normpath("/mypath/test/")]

    assert utils.utils.contains_parent_dir(path, dirs) is False


def test_deprecation_wrapper():
    """
    The DeprecationWrapper is not part of the public API

    We still need to test it.
    """

    a = utils.utils.DeprecationWrapper(
        DeprecatedClass,
        "DeprecatedClass",
        "NewClass",
        "1.9.0")
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        a()
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "deprecated" in str(w[-1].message)


class DeprecatedClass(object):
    """
    An empty class used for testing purposes.
    """
    pass



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import templates
from . import type_traits
from . import class_declaration
from . import class_declaration_traits
from . import class_traits
from . import traits_impl_details
from . import runtime_errors


class internal_type_traits(object):

    """small convenience class, which provides access to internal types"""
    # TODO: add exists function
    @staticmethod
    def get_by_name(type_, name):
        if class_traits.is_my_case(type_):
            cls = class_traits.get_declaration(type_)
            return type_traits.remove_declarated(
                cls.typedef(name, recursive=False).decl_type)
        elif class_declaration_traits.is_my_case(type_):
            cls = class_declaration_traits.get_declaration(type_)
            value_type_str = templates.args(cls.name)[0]
            ref = traits_impl_details.impl_details.find_value_type(
                cls.top_parent, value_type_str)
            if ref:
                return ref
            else:
                raise RuntimeError((
                    "Unable to find reference to internal " +
                    "type '%s' in type '%s'.") % (name, cls.decl_string))
        else:
            raise RuntimeError(
                ("Unable to find reference to internal type '%s'"
                 "in type '%s'.") % (name, type_.decl_string))


class smart_pointer_traits(object):

    """implements functionality, needed for convenient work with
    smart pointers"""

    @staticmethod
    def is_smart_pointer(type_):
        """returns True, if type represents instantiation of
        `boost::shared_ptr` or `std::shared_ptr`, False otherwise"""
        type_ = type_traits.remove_alias(type_)
        type_ = type_traits.remove_cv(type_)
        type_ = type_traits.remove_declarated(type_)
        if not isinstance(type_,
                          (class_declaration.class_declaration_t,
                           class_declaration.class_t)):
            return False
        if not (
                traits_impl_details.impl_details.is_defined_in_xxx(
                    'boost', type_) or
                traits_impl_details.impl_details.is_defined_in_xxx(
                    'std', type_)):
            return False
        return type_.decl_string.startswith('::boost::shared_ptr<') or \
            type_.decl_string.startswith('::std::shared_ptr<')

    @staticmethod
    def value_type(type_):
        """returns reference to `boost::shared_ptr` \
        or `std::shared_ptr` value type"""
        if not smart_pointer_traits.is_smart_pointer(type_):
            raise TypeError(
                'Type "%s" is not an instantiation of \
                boost::shared_ptr or std::shared_ptr' %
                type_.decl_string)
        try:
            return internal_type_traits.get_by_name(type_, "element_type")
        except runtime_errors.declaration_not_found_t:
            return _search_in_bases(type_)


class auto_ptr_traits(object):

    """implements functionality, needed for convenient work with
    `std::auto_ptr` pointers"""

    @staticmethod
    def is_smart_pointer(type_):
        """returns True, if type represents instantiation of
        `boost::shared_ptr`, False otherwise"""
        type_ = type_traits.remove_alias(type_)
        type_ = type_traits.remove_cv(type_)
        type_ = type_traits.remove_declarated(type_)
        if not isinstance(type_,
                          (class_declaration.class_declaration_t,
                           class_declaration.class_t)):
            return False
        if not traits_impl_details.impl_details.is_defined_in_xxx(
                'std', type_):
            return False
        return type_.decl_string.startswith('::std::auto_ptr<')

    @staticmethod
    def value_type(type_):
        """returns reference to `boost::shared_ptr` value type"""
        if not auto_ptr_traits.is_smart_pointer(type_):
            raise TypeError(
                'Type "%s" is not instantiation of std::auto_ptr' %
                type_.decl_string)
        try:
            return internal_type_traits.get_by_name(type_, "element_type")
        except runtime_errors.declaration_not_found_t:
            return _search_in_bases(type_)


def _search_in_bases(type_):
    """Implementation detail."""
    found = False
    for base_type in type_.declaration.bases:
        try:
            found = internal_type_traits.get_by_name(
                base_type.related_class, "element_type")
        except runtime_errors.declaration_not_found_t:
            pass
        if found:
            return found
    raise RuntimeError(
        ("Unable to find 'element_type' declaration '%s'"
         "in type '%s'.") % type_.decl_string)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

code = \
    """
class A {
public:
    virtual ~A() = 0;
    unsigned int a : 1;
    unsigned int unused : 31;
};
"""


def test_gccxml_10184():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse_string(code, config)
    global_ns = declarations.get_global_namespace(decls)
    assert global_ns.variable('a').bits == 1
    assert global_ns.variable('unused').bits == 31



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ['covariant_returns.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_is_same_function(global_ns):
    d = global_ns.class_('better_algorithm_t')
    b = global_ns.class_('algorithm_t')

    df = d.member_function('f')
    bf = b.member_function('f')

    assert id(df) != id(bf)
    assert declarations.is_same_function(df, bf) is True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import declarations


def __test_split_impl(decl_string, name, args):
    assert (name, args) == \
        declarations.templates.split(decl_string)


def __test_split_recursive_impl(decl_string, control_seq):
    assert control_seq == \
        list(declarations.templates.split_recursive(decl_string))


def __test_is_template_impl(decl_string):
    assert declarations.templates.is_instantiation(decl_string)


def test_split_on_vector():
    __test_is_template_impl("vector<int, std::allocator<int>>")

    __test_split_impl(
        "vector<int, std::allocator<int>>",
        "vector",
        ["int", "std::allocator<int>"])

    __test_split_recursive_impl(
        "vector<int, std::allocator<int>>",
        [("vector", ["int", "std::allocator<int>"]),
            ("std::allocator", ["int"])])


def test_split_on_string():
    __test_is_template_impl(
        "basic_string<char, std::char_traits<char>, std::allocator<char>>")

    __test_split_impl(
        "basic_string<char, std::char_traits<char>, std::allocator<char>>",
        "basic_string",
        ["char",
            "std::char_traits<char>",
            "std::allocator<char>"])


def test_split_on_map():
    __test_is_template_impl(
        "map<long int,std::vector<int, std::allocator<int>>," +
        "std::less<long int>, std::allocator<std::pair<const long int, " +
        "std::vector<int, std::allocator<int>>>>>")

    __test_split_impl(
        "map<long int,std::vector<int, std::allocator<int>>," +
        "std::less<long int>, std::allocator<std::pair<const long int, " +
        "std::vector<int, std::allocator<int>>>>>",
        "map",
        ["long int",
            "std::vector<int, std::allocator<int>>",
            "std::less<long int>",
            "std::allocator<std::pair<const long int, " +
            "std::vector<int, std::allocator<int>>>>"])


def test_join_on_vector():
    assert "vector<int, std::allocator<int>>" == \
        declarations.templates.join(
            "vector", ("int", "std::allocator<int>"))


def test_bug_is_tmpl_inst():
    assert declarations.templates.is_instantiation(
        "::FX::QMemArray<unsigned char>::setRawData") is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    # TODO: once gccxml is removed; rename this to something like
    # annotate_tester
    "attributes_castxml.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_attributes(global_ns):
    numeric = global_ns.class_('numeric_t')
    do_nothing = numeric.member_function('do_nothing')
    arg = do_nothing.arguments[0]
    assert "annotate(sealed)" == numeric.attributes
    assert "annotate(no throw)" == do_nothing.attributes
    assert "annotate(out)" == arg.attributes
    assert numeric.member_operators(name="=")[0].attributes is None


def test_attributes_thiscall():
    """
    Test attributes with the "f2" flag

    """

    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()

    config.flags = ["f2"]
    config.castxml_epic_version = 1

    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()

    numeric = global_ns.class_('numeric_t')
    do_nothing = numeric.member_function('do_nothing')
    arg = do_nothing.arguments[0]

    assert "annotate(sealed)" == numeric.attributes
    assert "annotate(out)" == arg.attributes

    assert "annotate(no throw)" == do_nothing.attributes
    assert numeric.member_operators(name="=")[0].attributes is None
